import os, sys, re, time, threading, _thread
import warnings
import configparser, inspect, getpass, traceback
from datetime import datetime, timedelta, timezone
from html.parser import HTMLParser
import mastodon
from mastodon import Mastodon, StreamListener
from configobj import ConfigObj

# TODO: Polish up sample bots for distribution (and real use!)
# TODO: Final pass on code quality and commenting
# TODO: Write up documentation

###
# Decorators
###

def reply(f):
    """ Mark f as a handler for mention notifications. """
    f.reply = True
    return f

def error_reporter(f):
    f.error_reporter = True
    return f

def interval(seconds):
    def wrapper(f):
        f.interval = seconds
        return f
    return wrapper

def schedule(**kwargs):
    def wrapper(f):
        if (not hasattr(f, "schedule")):
            f.schedule = [kwargs]
        else:
            f.schedule.append(kwargs)
        return f
    return wrapper

def hourly(minute=0):
    return schedule(minute=minute)

def daily(hour=0, minute=0):
    return schedule(hour=hour, minute=minute)

# Utilities

def total_seconds(dt):
    """Returns the total number of seconds in a timedelta."""
    return dt.seconds + dt.days * 24 * 60 * 60

def interval_next(f, t = datetime.now(), tLast = datetime.now()):
    """
    Calculate the number of seconds from now until the function should next run.
    This function handles both cron-like and interval-like scheduling via the
    following:
     ∗ If no interval and no schedule are specified, return 0
     ∗ If an interval is specified but no schedule, return the number of seconds
       from <t> until <interval> has passed since <tLast> or 0 if it's overdue.
     ∗ If a schedule is passed but no interval, figure out when next to run by
       parsing the schedule according to the following rules:
         ∗ If all of second, minute, hour, day_of_week/day_of_month, month, year
           are specified, then the time to run is singular and the function will
           run only once at that time. If it has not happened yet, return the
           number of seconds from <t> until that time, otherwise return -1.
         ∗ If one or more are unspecified, then they are treated as open slots.
           return the number of seconds from <t> until the time next fits within
           the specified constraints, or if it never will again return -1.
             ∗ Only one of day_of_week and day_of_month may be specified. if both
               are specified, then day_of_month is used and day_of_week is ignored.
         ∗ If all are unspecified treat it as having no schedule specified
     ∗ If both a schedule and an interval are specified, TODO but it should do
       something along the lines of finding the next multiple of interval from tLast 
       that fits the schedule spec and returning the number of seconds until then.

    NOTE: If the time until the next event is greater than an hour in the
    future, this function will return the number of seconds until the top of the
    next hour (1-3600). Be sure to continue checking until this function
    returns 0.
    """
    has_interval = hasattr(f, "interval")
    has_schedule = hasattr(f, "schedule")

    if (not has_interval and not has_schedule):
        return 0
    if (has_interval and not has_schedule):
        tNext = tLast + timedelta(seconds = f.interval)
        return max(total_seconds(tNext - t), 0)
    if (has_schedule): # and not has_interval):
        interval_min = 3600
        for s in f.schedule:
            interval = schedule_next(s, t) 
            if interval < interval_min:
                interval_min = interval
        return interval_min

def schedule_next(schedule, t):
    # Operate in day-of-month mode unless day-of-week is specified
    use_day_of_week = "day_of_week" in schedule

    spec = [schedule.get("hour", -1),
            (schedule.get("day_of_week", -1) if use_day_of_week
             else schedule.get("day_of_month", -1)),
            schedule.get("month", -1),
            schedule.get("year", -1)]
    now = [t.hour,
           t.weekday() if use_day_of_week else t.day,
           t.month,
           t.year]
    matches_hour_spec = all(spec[slot] == -1 or now[slot] == spec[slot] for slot in range(4))

    # Note that if decorator didn't specify second, we assume 0, we don't pattern match.
    spec = [schedule.get("second", 0),
            schedule.get("minute", -1)]
    now = [t.second,
           t.minute]
    now_s = 60 * now[1] + now[0]

    if matches_hour_spec:
        # Wait until minute:second matches time spec
        if spec[1] == -1:
            if now[0] < spec[0]: return spec[0] - now[0]
            elif now[0] > spec[0]: return 60 - (now[0] - spec[0])
            else: return 0
        else:
            spec_s = 60 * spec[1] + spec[0]
            if (now_s <= spec_s): return spec_s - now_s
            #else: fallthrough
    # else:
    # Wait until top of next hour
    return 3600 - now_s

class HTMLTextParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.text = ""
    def handle_data(self, data):
        self.text += data;

def html_strip_tags(html_str, linebreaks=None, lbchar="\n"):
    linebreaks = False if linebreaks == None else bool(linebreaks)
    if linebreaks:
        html_str = re.sub(r"<br([^>]*)>",
                          "{}<br\1>".format(lbchar),
                          html_str)
        html_str = re.sub(r"<p([^>]*)>",
                          "{}<p\1>".format(lbchar),
                          html_str)
    parser = HTMLTextParser()
    parser.feed(html_str)
    return parser.text

def get_mentions(status_dict, exclude=[]):
    """
    Given a status dictionary, return all people mentioned in the toot,
    excluding those in the list passed in exclude.
    """
    # Canonicalise the exclusion dictionary by lowercasing all names and
    # removing leading @'s
    for i, user in enumerate(exclude):
        user = user.casefold()
        if user[0] == "@":
            user = user[1:]

        exclude[i] = user

    users = [user["username"] for user in status_dict["mentions"]
             if user["username"].casefold() not in exclude]
    return users

def parse_list(list_str, sep=',', strip_separator_whitespace=True):
    list_str = list_str.strip()

    if list_str[0] == '[': list_str = list_str[1:]
    if list_str[-1] == ']': list_str = list_str[:-1]

    l = list_str.split(sep)
    if strip_separator_whitespace:
        l = [item.strip() for item in l]
    return l

class PineappleBot(StreamListener):
    """
    Main bot class
    We subclass StreamListener so that we can use it as its own callback
    """

    INITIALIZING = 0
    STARTING = 1
    RUNNING = 2
    STOPPING = 3

    class Config(dict):
        def __init__(self, bot, filename): 
            dict.__init__(self)
            self._filename = filename
            self._cfg = ConfigObj(filename, create_empty=True, interpolation="configparser")
            self._bot = bot
        def __getattr__(self, key): 
            if self[key]:
                return self[key]
            else:
                warnings.warn("The {} setting does not appear in {}. Setting the self.config value to None.".format(key, self._filename), 
                        RuntimeWarning, 1)
                return None
        def __setattr__(self, key, value): self[key] = value
        def __delattr(self, key): del self[key]

        def load(self, name=None):
            """ Load section <name> from the config file into this object,
            replacing/overwriting any values currently cached in here."""
            if (name != None):
                self._name = name

            self._cfg.reload()
            if (name not in self._cfg.sections): 
                self._bot.log("config", "Section {} not in {}, aborting.".format(self._name, self._filename))
                return False
            self._bot.log("config", "Loading configuration from {}".format(self._filename))
            if "DEFAULT" in self._cfg.sections:
                self.update(self._cfg["DEFAULT"])
            self.update(self._cfg[self._name])
            return True

        def save(self):
            """ Save back out to the config file. """
            self._cfg.reload()
            for attr, value in self.items():
                if attr[0] != '_' and not (attr in self._cfg["DEFAULT"] and self._cfg["DEFAULT"][attr] == value):
                    if isinstance(value, (list, tuple)):
                        value = ",".join([str(v) for v in value])
                    self._cfg[self._name][attr] = str(value)
            self._bot.log("config", "Saving configuration to {}...".format(self._filename))
            self._cfg.write()
            self._bot.log("config", "Done.")
            return True

    def __init__(self, cfgname, name=None, log_to_stderr=True, interactive=False, verbose=False):
        if (name is None): name = self.__class__.__name__
        self.name = name
        self.state = PineappleBot.INITIALIZING

        self.alive = threading.Condition()
        self.threads = []
        self.reply_funcs = []
        self.report_funcs = []

        self.mastodon = None
        self.account_info = None
        self.username = None
        self.default_visibility = None
        self.default_sensitive = None
        self.stream = None
        self.interactive = interactive
        self.verbose = verbose

        self.log_to_stderr = log_to_stderr
        self.log_name = self.name + ".log"
        self.log_file = open(self.log_name, "a")

        self.config = PineappleBot.Config(self, cfgname)
        self.init() # Call user init to initialize bot-specific properties to default values
        if not self.config.load(self.name): return
        if not self.login(): return

        self.startup()

    def log(self, id, msg):
        if (id == None): id = self.name
        else: id = self.name + "." + id
        ts = datetime.now()
        msg_f = "[{0:%Y-%m-%d %H:%M:%S}] {1}: {2}".format(ts, id, msg)

        if self.log_file.closed or self.log_to_stderr: print(msg_f, file=sys.stderr)
        elif not self.log_file.closed: print(msg_f, file=self.log_file)

    def startup(self):
        self.state = PineappleBot.STARTING
        self.log(None, "Starting {0} {1}".format(self.__class__.__name__, self.name))

        try:
            self.start()
        except Exception as e:
            self.log(None, "Fatal exception: {}\n{}".format(repr(e), traceback.format_exc()))
            return
        
        def interval_threadproc(f):
            self.log(f.__name__, "Started")
            t = datetime.now()
            tLast = t
            while (True):
                self.alive.acquire()
                t = datetime.now()
                interval = interval_next(f, t, tLast)

                if (interval == 0):
                    try:
                        f()
                    except Exception as e:
                        error = "Exception encountered in @interval function: {}\n{}".format(repr(e), traceback.format_exc())
                        self.report_error(error, f.__name__)

                    t = datetime.now()
                    interval = interval_next(f, t, t)

                if self.verbose: self.log(f.__name__ + ".debug", "Next wait interval: {}s".format(interval))
                tLast = t
                self.alive.wait(max(interval, 1))
                if (self.state == PineappleBot.STOPPING):
                    self.alive.release()
                    self.log(f.__name__, "Shutting down")
                    return 0
                else:
                    self.alive.release()

        for fname, f in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(f, "interval") or hasattr(f, "schedule"):
                t = threading.Thread(args=(f,), target=interval_threadproc)
                t.start()
                self.threads.append(t)

            if hasattr(f, "reply"):
                self.reply_funcs.append(f)

            if hasattr(f, "error_reporter"):
                self.report_funcs.append(f)

        if len(self.reply_funcs) > 0:
            self.stream = self.mastodon.stream_user(self, async=True)

        credentials = self.mastodon.account_verify_credentials()
        self.account_info = credentials
        self.username = credentials["username"]
        self.default_visibility = credentials["source"]["privacy"]
        self.default_sensitive = credentials["source"]["sensitive"]

        self.state = PineappleBot.RUNNING
        self.log(None, "Startup complete.")

    def shutdown(self):
        self.alive.acquire()
        self.state = PineappleBot.STOPPING
        self.log(None, "Stopping {0} {1}".format(self.__class__.__name__, self.name))
        self.alive.notify_all()
        self.alive.release()

        if self.stream: self.stream.close()

        self.stop()
        self.config.save()

        self.log_file.close()

    def login(self):
        if self.interactive and not self.interactive_login(): 
            self.log("api", "Interactive login failed, exiting.")
            return False
        elif "domain" not in self.config:
            self.log("api", "No domain set in config and interactive = False, exiting.")
            return False
        elif "client_id" not in self.config:
            self.log("api", "No client id set in config and interactive = False, exiting.")
            return False
        elif "client_secret" not in self.config:
            self.log("api", "No client secret set in config and interactive = False, exiting.")
            return False
        elif "access_token" not in self.config:
            self.log("api", "No access key set in config and interactive = False, exiting.")
            return False

        self.mastodon = Mastodon(client_id = self.config.client_id, 
                                  client_secret = self.config.client_secret, 
                                  access_token = self.config.access_token, 
                                  api_base_url = self.config.domain)
                                  #debug_requests = True)
        return True

    def interactive_login(self):
        try:
            if (not hasattr(self, "domain")):
                domain = input("{0}: Enter the instance domain [mastodon.social]: ".format(self.name))
                domain = domain.strip()
                if (domain == ""): domain = "mastodon.social"
                self.config.domain = domain
            # We have to generate these two together, if just one is 
            # specified in the config file it's no good.
            if (not hasattr(self, "client_id") or not hasattr(self, "client_secret")):
                client_name = input("{0}: Enter a name for this bot or service [{0}]: ".format(self.name))
                client_name = client_name.strip()
                if (client_name == ""): client_name = self.name
                self.config.client_id, self.config.client_secret = Mastodon.create_app(client_name, 
                        api_base_url="https://"+self.config.domain)
                # TODO handle failure
                self.config.save()
            if (not hasattr(self, "access_token")):
                email = input("{0}: Enter the account email: ".format(self.name))
                email = email.strip()
                password = getpass.getpass("{0}: Enter the account password: ".format(self.name))
                try:
                    mastodon = Mastodon(client_id = self.config.client_id,
                                        client_secret = self.config.client_secret,
                                        api_base_url = "https://"+self.config.domain)
                    self.config.access_token = mastodon.log_in(email, password)
                    self.config.save()
                except ValueError as e:
                    self.log("login", "Could not authenticate with {0} as '{1}':"
                             .format(self.config.domain, email))
                    self.log("login", str(e))
                    self.log("debug", "using the password {0}".format(password))
                    return False
            return True
        except KeyboardInterrupt:
            return False

    @error_reporter
    def default_report_handler(self, error):
        if self.mastodon and "admin" in self.config:
            self.mastodon.status_post(("@{} ERROR REPORT from {}:\n{}"
                                       .format(self.config.admin, self.name, error))[:500],
                                      visibility="direct")

    def report_error(self, error, location=None):
        """Report an error that occurred during bot operations. The default
        handler tries to DM the bot admin, if one is set, but more handlers can
        be added by using the @error_reporter decorator."""
        if location == None: location = inspect.stack()[1][3]
        self.log(location, error)
        for f in self.report_funcs:
            f(error)

    def on_notification(self, notif):
        self.log("debug", "Got a {} from {} at {}".format(notif["type"], notif["account"]["username"], notif["created_at"]))
        if (notif["type"] == "mention"):
            for f in self.reply_funcs:
                try:
                    f(notif["status"], notif["account"])
                except Exception as e:
                    error = "Fatal exception: {}\n{}".format(repr(e), traceback.format_exc())
                    self.report_error(error, f.__name__)

    def on_close(self):
        # Attempt to re-open the connection, since this is usually the result of
        # the server just dropping the connection for one reason or another.
        self.log(None, "Dropped streaming connection to {}".format(self.config.domain))
        self.stream.close()
        old_timeout = self.mastodon.request_timeout
        # Don't wait forever for the instance to respond to the /api/v1/instance endpoint inside Mastodon.py
        #self.mastodon.request_timeout = (10, 10)
        print("thing1")
        try:
            self.mastodon.instance()
        except mastodon.Mastodon.MastodonNetworkError as e:
            self.log("Instance appears to have gone down")

        print("thing2")
        wait = 10
        while(True):
            try:
                self.log(None, "Attempting to reinitialize in {}s...".format(wait))
                time.sleep(wait)
                self.stream = self.mastodon.stream_user(self, async=True)
                # Call the instance API first, so that we don't get stuck in stream_user
                #  (timeout doesn't work there for some reason)
                self.mastodon.instance()

                # If we get here without erroring, success
                self.mastodon.request_timeout = old_timeout
                self.log(None, "Successfully reinitialized streaming connection.")
                break
            except mastodon.Mastodon.MastodonNetworkError:
                self.log(None, "Timed out")
                wait *= 2
                continue

    def get_reply_visibility(self, status_dict):
        """Given a status dict, return the visibility that should be used.
        This behaves like Mastodon does by default.
        """
        # Visibility rankings (higher is more limited)
        visibility = ("public", "unlisted", "private", "direct")

        default_visibility = visibility.index(self.default_visibility)
        status_visibility = visibility.index(status_dict["visibility"])

        return visibility[max(default_visibility, status_visibility)]

    # defaults, should be replaced by concrete bots with actual implementations
    # (if necessary, anyway)
    def init(self):
        pass
    def start(self):
        pass
    def stop(self):
        pass

# Exceptions

class ConfigurationError(Exception):
    pass



from ananas import PineappleBot, interval, ConfigurationError

class AnnounceBot(PineappleBot):
    """Bot that regularly checks approved users' feeds for posts with a
    particular hashtag and boosts them when found. Good for automated
    announcement bots which draw content from admins' feeds."""

    def init(self):
        self.config.hashtag = "announcement"
        self.config.allow_list = "admin"
        self.config.last_seen = 0

    def start(self):
        # Normalize the hashtag value
        if self.config.hashtag[0] == "#": self.config.hashtag = self.config.hashtag[1:]
        self.config.hashtag = self.config.hashtag.lower()

        # Tolerate not having actual lists in the underlying config
        if isinstance(self.config.allow_list, str):
            self.config.allow_list = [self.config.allow_list]

        if isinstance(self.config.last_seen, str):
            self.config.last_seen = [self.config.last_seen]

        # Make sure last seen ids are actually ints
        self.config.last_seen = [int(id) for id in self.config.last_seen]

        # Grab the actual user dicts for the usernames listed in the config
        self.users = []
        for username in self.config.allow_list:
            users = self.mastodon.account_search(username, 1)
            if len(users) == 0 or users[0].username.lower() != username:
                raise ConfigurationError("Unknown user {} in allow list".format(username))
            self.users.append(users[0])

        assert len(self.config.allow_list) == len(self.users)

        # Make sure the last seen array is the right length
        if len(self.config.last_seen) != len(self.config.allow_list):
            raise ConfigurationError("There must be the same number of last seen IDs as allowed users")

        # Run initial update immediately
        self.update()

    # Check every 1 minute
    @interval(1 * 60)
    def update(self):
        for i, user in enumerate(self.users):
            lowest_seen_this_time = self.config.last_seen[i]
            
            while (True):
                posts = self.mastodon.account_statuses(user, since_id=lowest_seen_this_time+1, exclude_replies=True)
                for post in posts:
                    post_id = post.id

                    if post.reblog:
                        post = post.reblog

                    for tag in post.tags:
                        if tag.name == self.config.hashtag:
                            self.log("debug", "BOOST: {}".format(post.id))
                            #mastodon.status_reblog(post)
                            self.config.last_seen[i] = post_id
                            break

                if self.config.last_seen[i] == lowest_seen_this_time: break
                else: lowest_seen_this_time = self.config.last_seen[i]





#!/usr/bin/env python3
import os, sys, signal, argparse, configparser, traceback, time
from contextlib import closing
from ananas import PineappleBot
import ananas.default

# Add the cwd to the module search path so that we can load user bot classes
sys.path.append(os.getcwd())

bots = []

def shutdown_all(signum, frame):
    for bot in bots:
        if bot.state == PineappleBot.RUNNING: bot.shutdown()
    sys.exit("Shutdown complete")

def main():
    parser = argparse.ArgumentParser(description="Pineapple command line interface.", prog="ananas")
    parser.add_argument("config", help="A cfg file to read bot configuration from.")
    parser.add_argument("-v", "--verbose", action="store_true", help="Log more extensive messages for e.g. debugging purposes.")
    parser.add_argument("-i", "--interactive", action="store_true", help="Use interactive prompts for e.g. mastodon login")
    args = parser.parse_args()

    prog = sys.argv[0]

    cfg = configparser.ConfigParser()
    try: cfg.read(args.config)
    except FileNotFoundError:
        sys.exit("Couldn't open '{}', exiting.".format(args.config))

    for bot in cfg:
        if bot == "DEFAULT": continue
        if not "class" in cfg[bot]:
            print("{}: no class specified, skipping {}.".format(prog, bot))
            continue

        botclass = cfg[bot]["class"]
        module, _, botclass = botclass.rpartition(".")
        if module == "":
            print("{}: no module given in class name '{}', skipping {}.".format(prog, botclass, bot))

        try:
            exec("from {0} import {1}; bots.append({1}('{2}', name='{3}', interactive={4}, verbose={5}))"
                    .format(module, botclass, args.config, bot, args.interactive, args.verbose))
        except ModuleNotFoundError as e:
            print("{}: couldn't load module {}, skipping {}.".format(prog, module, bot))
            continue
        except Exception as e:
            print("{}: fatal exception loading bot {}: {}\n{}".format(prog, bot, repr(e), traceback.format_exc()))
            continue
        except KeyboardInterrupt:
            sys.exit()

    signal.signal(signal.SIGINT, shutdown_all)
    signal.signal(signal.SIGABRT, shutdown_all)
    signal.signal(signal.SIGTERM, shutdown_all)

    try:
        while(True): time.sleep(60)
    except KeyboardInterrupt:
        shutdown_all(None, None)

if __name__ == "__main__":
    main()



import random, string, itertools, traceback
from more_itertools import peekable
from ananas import PineappleBot, reply, html_strip_tags

def peek(gen): return gen.peek()

# DICE PARSING
# Formats we want to accept:
#  @roll [#]d#[±#]
#  @roll #d#d#[±#]
#  @roll #d#k#[±#]
#  @roll #x[above]
# We also want to accept any amount of whitespace (but not other
# characters) between portions of the roll:
#  @roll 3 d20 + 2
#  @roll 3d20 k 4 - 2
#
# We'll accept multiple rolls, but only with semicolons, commas, or
# linebreaks beween them
#  @roll 2d20, 2d6
#
# Finally, we'll accept simple arithmetic expressions of rolls:
#  @roll 2d20 + 2d6 + 3
#
# The syntax tree is as follows:
# A single dice roll is a tuple:
#  ('r', <num sides>, <num dice>, ['d'/'k', <num to drop/keep>])
# An arithmetic expression is a tuple:
#  (<op>, <lhs>, <rhs>)
# Constants appearing in arithmetic expressions are:
#  ('c', <constant>)
# A series of dice rolls are a list:
# [(roll1), ...(rolln)]
#
# Example:
#  3d20k1 + d8 + 1
# becomes
#  ('+', ('r', 3, 20, 'k', 1) ('+', ('r', 1, 8), ('c', 1)))

ops = "+-*x"

def parse_dice(text):
    class AbortedParseError(Exception): pass

    def expect(gen, pred): 
        val,ws = gen.peek()
        if (not pred(val,ws)): raise AbortedParseError("parse got unexpected value {} in {}".format(val, pred.__code__))
        return next(gen)[0]

    def tokenizer():
        s = ""
        ws = False
        for c in text:
            if c in string.digits: s += c
            else:
                if len(s) > 0:
                    yield int(s), ws
                    s = ""
                    ws = False
                if c in string.whitespace: ws = True; continue
                elif c in ops+"dk;,\n": yield c, ws; ws = False
                elif c == '\U0001F4AF': yield 100, ws; ws = False
                else: yield 'z', ws; ws = False
        if len(s) > 0:
            yield int(s), ws
        return

    def parse_roll_list(tokens):
        rolls = []
        try:
            roll = parse_roll_add_expr(tokens)
            if (roll[0] != 'c'):
                rolls.append(roll)
        except StopIteration: return rolls
        except AbortedParseError as e: pass #print(repr(e))
        except ValueError as e: pass #print(repr(e))

        while True:
            try:
                next(tokens)
                roll = parse_roll_add_expr(tokens)
                if (roll[0] != 'c'):
                    rolls.append(roll)
            except StopIteration: break
            except AbortedParseError as e: continue #print(repr(e)); continue
            except ValueError as e: continue #print(repr(e)); continue

        return rolls

    # 2 * 3 x 4
    # (* 2 3)
    # (x (* 2 3) 4)
    # (* 2 (x 3 4))

    # 2 x 3 x 4
    # (x 2 3)
    # (x (x 2 3) 4)

    def parse_roll_mul_expr(tokens):
        lhs = parse_roll(tokens)
        while True:
            try: 
                if (str(peek(tokens)[0]) not in "*x"): break
                op = expect(tokens, lambda t,ws: t in "*x")
                rhs = parse_roll(tokens)
                lhs = (op, lhs, rhs)
            except AbortedParseError: return lhs
            except StopIteration: return lhs
        return lhs

    def parse_roll_add_expr(tokens):
        lhs = parse_roll_mul_expr(tokens)
        while True:
            try: 
                if (peek(tokens)[0] not in "+-"): break
                op = expect(tokens, lambda t,ws: t in "+-")
                rhs = parse_roll_mul_expr(tokens)
                lhs = (op, lhs, rhs)
            except AbortedParseError: return lhs
            except StopIteration: return lhs
        return lhs

    def parse_roll(tokens):
        p,ws = peek(tokens)
        if (p == 'd'): c = -1
        else: c = expect(tokens, lambda t,ws: isinstance(t, int))

        try:
            d = expect(tokens, lambda t,ws: t == 'd')
            sides = expect(tokens, lambda t,ws: isinstance(t, int) and not ws)
        except AbortedParseError:
            return ('c', c)
        except StopIteration: 
            if c < 0: raise ValueError() # lone 'd'
            return ('c', c)

        if (c < 0): return ('r', 1, sides)

        try:
            dk = expect(tokens, lambda t,ws: t == 'd' or t == 'k')
            num = expect(tokens, lambda t,ws: isinstance(t, int))
        except StopIteration: 
            return ('r', c, sides)
        except AbortedParseError:
            return ('r', c, sides)

        return ('r', c, sides, dk, num)

    return parse_roll_list(peekable(tokenizer()))

# Syntax tree visitors
# They only do singular roll expressions since the calling code is likely to
# want to unwrap the rolls and handle them one by one

def spec_dice(spec):
    """ Return the dice specification as a string in a common format """
    if spec[0] == 'c': 
        return str(spec[1])
    elif spec[0] == 'r':
        r = spec[1:]
        s = "{}d{}".format(r[0], r[1])
        if len(r) == 4 and ((r[2] == 'd' and r[3] < r[0]) or (r[2] == 'k' and r[3] > 0)):
            s += "{}{}".format(r[2], r[3])
        return s
    elif spec[0] in ops:
        return "{} {} {}".format(spec_dice(spec[1]), spec[0], spec_dice(spec[2]))
    else: raise ValueError("Invalid dice specification")

def roll_dice(spec):
    """ Perform the dice rolls and replace all roll expressions with lists of
    the dice faces that landed up. """
    if spec[0] == 'c': return spec
    if spec[0] == 'r':
        r = spec[1:]
        if len(r) == 2: return ('r', perform_roll(r[0], r[1]))
        k = r[3] if r[2] == 'k' else -1 
        d = r[3] if r[2] == 'd' else -1
        return ('r', perform_roll(r[0], r[1], k, d))
    if spec[0] == "x":
        c = None
        roll = None
        if spec[1][0] == "c": c = spec[1]
        elif spec[1][0] == "r": roll = spec[1]
        if spec[2][0] == "c": c = spec[2]
        elif spec[2][0] == "r": roll = spec[2]

        if (c == None or roll == None):
            return ('*', roll_dice(spec[1]), roll_dice(spec[2]))
        else:
            if (c[1] > 50):
                raise SillyDiceError("I don't have that many dice!")
            return ("x", [roll_dice(roll) for i in range(c[1])])
    if spec[0] in ops:
        return (spec[0], roll_dice(spec[1]), roll_dice(spec[2]))
    else: raise ValueError("Invalid dice specification")

def sum_dice(spec):
    """ Replace the dice roll arrays from roll_dice in place with summations of
    the rolls. """
    if spec[0] == 'c': return spec[1]
    elif spec[0] == 'r': return sum(spec[1])
    elif spec[0] == 'x':
        return [sum_dice(r) for r in spec[1]]
    elif spec[0] in ops:
        return (spec[0], sum_dice(spec[1]), sum_dice(spec[2]))
    else: raise ValueError("Invalid dice specification")

def eval_dice(spec):
    if spec[0] == 'c': return spec[1]
    elif spec[0] == 'r': return sum(spec[1])
    elif spec[0] == 'x': return sum(eval_dice(r) for r in spec[1])
    elif spec[0] == '+': return eval_dice(spec[1]) + eval_dice(spec[2])
    elif spec[0] == '-': return eval_dice(spec[1]) - eval_dice(spec[2])
    elif spec[0] == '*': return eval_dice(spec[1]) * eval_dice(spec[2])
    else: raise ValueError("Invalid dice specification")

def visit_dice(d):
    if d[0] == 'c': return str(d[1])
    elif d[0] == 'r':
        s = ""
        for v in d[1]: s += "[{}]".format(v)
        return s
    elif d[0] == 'x':
        s = ""
        return " + ".join([visit_dice(v) for v in d[1]])
        return s
    else:
        return "{} {} {}".format(visit_dice(d[1]), d[0], visit_dice(d[2]))

def visit_sum_dice(d):
    if isinstance(d, int): return str(d)
    elif isinstance(d, list): return " + ".join([str(r) for r in d])
    else: return "{} {} {}".format(visit_sum_dice(d[1]), d[0], visit_sum_dice(d[2]))

# Roll <dice> <sides>-sided dice
# If <keep> is specified, only return the top <keep>
# If <drop> is specified, return all but the bottom <drop>
def perform_roll(dice=1, sides=6, keep=-1, drop=-1):
    r = []
    rolls = []
    if sides == 0: 
        raise SillyDiceError("I don't have any zero-dimensional constructs but when I find one, I'll get back to you.")
    if dice > 50:
        raise SillyDiceError("I don't have that many dice!")
    if sides > 1000:
        raise SillyDiceError("I rolled the sphere and it rolled off the table.")
    for i in range(dice):
        roll = random.randint(1, sides)
        r.append(roll)
        rolls.append((i, roll))
    rolls.sort(key=lambda roll: -roll[1])

    if keep > 0:
        rolls = rolls[0:keep]
    elif drop > 0 and drop < dice:
        rolls = rolls[0:-drop]
    elif drop == -1 and keep == -1: pass
    else:
        raise SillyDiceError("Whoops, dropped all the dice")

    if len(rolls) < len(r):
        r = [roll for i,roll in enumerate(r) if (i, roll) in rolls]

    return r

class DiceBot(PineappleBot):
    @reply
    def handle_roll(self, mention, user):
        raw = html_strip_tags(mention["content"]) 
        username = user["acct"]

        self.log("handle_roll", "Parsing dice in '{}' from @{}" .format(raw, username))
        message = ""
        try:
            rolls = parse_dice(raw)
        except Exception as e:
            self.report_error("{}\n{}".format(repr(e), traceback.format_exc()))
            rolls = []

        if len(rolls) == 0:
            rolls = [('r', 1, 6)]
            message = "I'm confused, so I'm just going to roll a d6.\n"
            return # TODO cheeky message?
        else:
            self.log("debug", "rolls: {}".format(rolls))

        for i, r in enumerate(rolls):
            try:
                #r = fixup_tree(r)
                line = "Rolling {}: ".format(spec_dice(r))
                dice = roll_dice(r)
                expr = sum_dice(dice)
                sum = eval_dice(dice)
                line += "{} = {}".format(visit_dice(dice), visit_sum_dice(expr))
                if (not isinstance(expr, int)):
                    line += " = {}".format(sum)
                message = line + "\n"
            except SillyDiceError as e:
                message += str(e) + "\n"

        self.mastodon.status_post("@{}\n{}".format(username, message),
                in_reply_to_id = mention["id"],
                visibility = mention["visibility"])


class SillyDiceError(Exception): pass
